/* include files for Motif and X Window functions */
#include <Xm/Separator.h>
#include <Xm/ToggleBG.h>
#include <Xm/MainW.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/ToggleB.h>
#include <Xm/CascadeB.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/Label.h>
#include <Xm/ScrollBar.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <unistd.h>
#ifndef dsc_h
	#define dsc_h

   /****************************************************************************
	*                        Handy Definitions                                 *
	****************************************************************************/
	#define PI 3.1415926535897932384626433
	#define TRUE  1
	#define FALSE 0
	#define FATAL 0
	#define NON_FATAL 1
	#define ROTATE_Y_X 0
	#define ROTATE_Y_Z 1
	#define TRANSLATE 2
	#define SCALE_SLAB 3
	#define DISTANCE 0
	#define ANGLE 1
	#define DIHEDRAL 2
	#define MAX_PICKED_ATOMS 100
	#define SBincrement 1        /* Scrollbar increment */
	#define SBmax 180            /* Scrollbar maximium value */

   /****************************************************************************
	*                        Structure Definitions                             *
	****************************************************************************/
    typedef struct { float xyz[3]; float length; } Vector;
	typedef struct {
		float x,y,z;
		char name[10];
		float red,green,blue;
		float dx,dy,dz;
		int serial;
		int isslabed;
		int ispicked;
	} draw_ATOM;
	typedef struct { void *first,*second; } draw_BOND;
	typedef struct { void *what, *next; int id; } list_link;
	typedef struct { int hlv, vlv; } SBList;
	typedef struct { int num, show, calc; draw_ATOM *atoms[MAX_PICKED_ATOMS]; } PL;

   /****************************************************************************
	*                      Global Function Declarations                        *
	****************************************************************************/
	void GrammpError(char *, int);
	void dscMainLoop(XtAppContext);
	void dscExec(XtAppContext);
	void dscDispatchEvent(XEvent *);
	void dscIdleFunc(void (*)(void));
	void dscMouseMoveFunc(void (*)(XEvent *));
	int dscSystem(char *);
	void dscDestroyWidget(Widget, Widget, XtPointer);
	ssize_t ReadFromGrfx(void *, size_t);
	ssize_t ReadFromAmmp(void *, size_t);
	ssize_t WriteToGrfx(void *, size_t);
	ssize_t WriteToAmmp(void *, size_t);
	void ScaleMolecule(float *);
	void CenterMolecule(void);
	void LookDownAxis(Widget, char *);
	void SendToScreen(Widget);
	void Rotate(char, float);
	void Translate(char, float);


   /****************************************************************************
	*                        Global variables                                  * 
	****************************************************************************/
	Widget VSB;                         /* Vertical scroll bar */
	Widget HSB;                         /* Horizontal scroll bar */
	int LEFT_BUTTON_DOWN;               /* left mouse button down? */
	draw_ATOM *last_picked_draw_ATOM;
	list_link *first_draw_ATOM;
	list_link *last_draw_ATOM;
	list_link *first_draw_BOND;
	list_link *last_draw_BOND;
	int ammp_to_grfx_pipe[2];          /* pipe from ammp to grfx processes */
	int grfx_to_ammp_pipe[2];          /* pipe from grfx to ammp processes */
	int Autocenter;                    /* automatically center molecule? */
	int Autoscale;                     /* automatically scale molecule? */
	int OptionMode;                    /* scrollbar option mode */
	char MolecularColor[256];          /* describes molecular color scheme */
	float draw_winx, draw_winy;        /* defines size of rendering window */
	static int VisualArgs[] = {        /* arguments for OpenGL rendering context */
		GLX_DOUBLEBUFFER, GLX_RGBA,
		GLX_DEPTH_SIZE, 16,
		GLX_RED_SIZE, 1, 
		GLX_GREEN_SIZE, 1,
		GLX_BLUE_SIZE, 1,
		None
	};
	static int *snglBuf = &VisualArgs[1];
	Display *display;                  /* pointer to X display structure */
	GLboolean made_current;            /* Is the GL display bound to a X window */
	Widget Glxarea;                    /* Drawing Area widget which supports OpenGL output */
	XVisualInfo *vi;                   /* stores information about X context */
	int LOADING;                       /* boolean for whether molecule is still loading */
	int show_tethers;                  /* boolean for whether tethers should be shown */
	int show_noels;                    /* boolean for whether noels should be shown */
	int show_violations;               /* boolean for whether violations should be shown */
	static void (*IdleFunc)(void) = FALSE;
	static void (*MouseMoveFunc)(XEvent *) = FALSE;
	SBList SB[10];                     /* structures for scroll bar information */
	PL PickList[10];                   /* structures for geometric calculations */
	float draw_center[3];              /* coordinates of the drawing window center */
#endif
