/* richard.c
*
*  thermodynamic path integrals using the
*  Feynmann Weiner Kac approach (actually the Wiener version)
*
*
*  basically:
*
* follow the integral of the exp(lagrangian)
* over the path
*
*  exp( - ( 0.5 m Delta((v )^2)dt - Delta(V) dt) / (kT dt) )
*  where v is velocity, V potential energy and everything
* else is obvious
*
* for each atom therefor:
*
*  occupancy *=  exp( -delta(V)/kT ) *exp( 0.5 m Delta(v^2)/kT)
*  where the potential change is global, but the 
*   velocity changes are local 
*
*  the total state occupancy is therefore:
*
*  sum( atom occupancy) = exp( -delta(V)/kT)*sum( exp(0.5....) )
*
*  This can be seen either as a limit in small time steps
*  or a global potential (i.e. function of all atoms at once)
*
*  however in the absense of atom creation or anihilation
*  the sum( exp(0.5 ...)) must be constant (=Natom)
*  (if later we wish to use it this is a constraint)
*  (i.e. the integral over all paths of the free particle propagator
*   must be the total number of free particles)
*
* >>>>  Therefore in order to study path integral averages over
* >>>>  the system  it is only necessary to track 
* >>>>   exp( -delta(V)/kT)  as a product from step to step.
*
*  nifty?
*/
#define ANSI 1
/* misc includes - ANSI and some are just to be safe */
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#ifdef ANSI
#include <stdlib.h>
#endif
#include "ammp.h"


int richard( op,echo, vfs,ffs,nfs,who,niter,p1,p2,p3)
FILE *op;
int echo;
int (*vfs[])(),(*ffs[])(),nfs;
int niter;
char *who;
float p1,p2,p3;
{
	int pac(),pacpac(),tpac(),hpac(),ptpac(),verlet();
	int (*touse)();
	int i;
	float V,Vold,T;
	int set_f_variable();
	int numatm, a_number();
	float get_f_variable(),o;
/* figure out which  function to use */
	touse = NULL;
	if( strcmp( who,"pac") == 0){  touse = pac;}
	else if( strcmp(who,"pacpac") == 0){touse = pacpac;}
	else if( strcmp(who,"tpac") == 0){touse = tpac;}
	else if( strcmp(who,"hpac") == 0){touse = hpac;}
	else if( strcmp(who,"ptpac") == 0){touse = ptpac;}
	else if( strcmp(who,"verlet") == 0){touse = verlet;}
	if( touse == NULL ){
	aaerror("usage richard <pac,pacpac,tpac,hpac,ptpac> nstep \n");
	return 1;
	}

	numatm = a_number();
	T = p2; if( touse == ptpac) T = p3; if( T <= 0.00001) T = 300.;
	V= 0.;
	for( i=0; i<nfs; i++)
		(*vfs[i])(&V,0.);

	for( i=0; i< 1; i++)
	{
		if( touse == pac || touse == pacpac || touse == verlet)
		{ (*touse)(ffs,nfs,niter,p1);}
			else if( touse == tpac)
			{(*touse)(ffs,nfs,niter,p1,p2);}
			else if( touse == hpac)
			{ (*touse)(ffs,vfs,nfs,niter,p1,p2);}
			else 
			{ (*touse)(ffs,nfs,niter,p1,p2,p3);}
	}/* end of for(i) */
	Vold = V;
	V= 0.;
	for( i=0; i<nfs; i++)
		(*vfs[i])(&V,0.);
	o = get_f_variable("occup");
	if( o <= 0.) o = 1.;
	o *= exp( -(V-Vold)/(1.987e-3*T*numatm));
	set_f_variable( "occup",o);
	return 0;
}/* end of richard */
