/*
* tncnewt.c
*
* truncated newtons method
* optimiser routine for AMMP
*
* theory is to use a finite difference appx to the
* hessian step product and find the optimal step vector
* that way.
*
* basically this is fixed point iteration for the
* difference in gradients being colinear and 
* descent direction along the step vector.
* 
*/
/*
*  copyright 1992 Robert W. Harrison
*  
*  This notice may not be removed
*  This program may be copied for scientific use
*  It may not be sold for profit without explicit
*  permission of the author(s) who retain any
*  commercial rights including the right to modify 
*  this notice
*/

#define ANSI 1
#include <stdio.h>
#include <math.h>
#ifdef ANSI
#include <stdlib.h>
#endif
#include "ammp.h"

/* function tncnewt()
* truncated newtons method optimiser
*
* passed a set of functions vfs for the potential
* passed a set of functions ffs for the force
*  how many functions  nfs
*  how many steps to try
*  when to stop
*/
int tncnewt( vfs,ffs,nfs, nstep,dtoler,toler )
int nfs,(*vfs[])(),(*ffs[])();
int nstep;
float toler,dtoler;
{
	int i,j,ifs,istep;
	float vb,vt,vto,vtk,alpha,atop;
	float dfx,dfy,dfz;
	float linmin();
	float a_max_f(),a_max_d(),lam;
	ATOM *ap,*a_next();
/*  do at most nstep steps */
	printf(" entering tncnewt \n");
	if( dtoler < 1.e-3) dtoler = .001;
	for( istep=0; istep< nstep ; istep++)
	{
	vb = 0.;
	for( ifs = 0; ifs < nfs; ifs++)
		(*vfs[ifs])( &vb,  0.);
	a_f_zero();
	for( ifs = 0; ifs < nfs; ifs++)
		(*ffs[ifs])( 0.);
	lam = a_max_f();
	printf(" vb, maxf %f %f \n",vb,(lam));
	if( lam <= toler) return 1;
	if( lam <= 1. )
	{ a_ftodx(1.,0.); } else { a_ftodx( 1./sqrt(lam),0. ); }
	a_ftovx(1.,0.);
/* initial iteration to find the step vector */
/* using a small number of trial searches
* rather than the largish number in tnpack
*
* but not checking for convergence 
*/
	a_v_zero();
	for( i=0; i< 5; i++)
	{
	a_f_zero();
	for( ifs = 0; ifs < nfs; ifs++)
		(*ffs[ifs])( dtoler);
	ap = a_next(-1);   
/* form the Hd product and alpha */
	alpha = 0.; atop = 0.; 
	while( 1)
	{
	if( ap == NULL) break;
/* this is the product
* well use it on the fly 
	ap->fx = (ap->fx-ap->dx)/dtoler;
	ap->fy = (ap->fy-ap->dy)/dtoler;
	ap->fz = (ap->fz-ap->dz)/dtoler;
*/
	dfx = (ap->fx-ap->dx)/dtoler;
	dfy = (ap->fy-ap->dy)/dtoler;
	dfz = (ap->fz-ap->dz)/dtoler;
	alpha += ap->dx*dfx;
	alpha += ap->dy*dfy;
	alpha += ap->dz*dfz;
	atop += (dfx -ap->fx)*ap->dx;
	atop += (dfy -ap->fy)*ap->dy;
	atop += (dfz -ap->fz)*ap->dz;
	if( ap == ap->next) break;	
	ap = ap->next;
	}
	lam = sqrt(a_max_f());if( lam < 1.) lam = 1.; 
	ap = a_next(-1);   
	while( 1)
	{
	if( ap == NULL) break;
	ap->vx += atop/alpha * ap->dx; ap->dx = ap->fx/lam ;
	ap->vy += atop/alpha * ap->dy; ap->dy = ap->fy/lam ;
	ap->vz += atop/alpha * ap->dz; ap->dz = ap->fz/lam ;
	if( ap == ap->next) break;	
	ap = ap->next;
	}
	}
	ap = a_next(-1);   
	while( 1)
	{
	if( ap == NULL) break;
	ap->dx  =  ap->vx;
	ap->dy  =  ap->vy;
	ap->dz  =  ap->vz;
	if( ap == ap->next) break;	
	ap = ap->next;
	}


	lam = linmin( vfs,nfs, .01);
	if( lam < 1.e-6) 
	{
	a_f_zero();
	for( ifs = 0; ifs < nfs; ifs++)
		(*ffs[ifs])( 0.);
	lam = sqrt(a_max_f()); if( lam < 1.) lam = 1.; 
	a_ftodx(1./lam, 0.);
	lam = linmin(vfs,nfs,.01);
	if( lam < 1.e-8) return 1;
	vt = 0.;
	for( ifs = 0; ifs < nfs; ifs++)
		(*vfs[ifs])( &vt,lam);
	printf(" steep sub-iter  %f %f\n",vt,vb);
	if( vt > vb) return 0;
	}
	a_inc_d( lam );
	}
	return 0;
}
